local utils = require("bliss.utils")

--==============================================================================
-- Helpers

local akp = {}

function akp.is_valid_file(data)
    m:seek(0) -- restart from beginning
    local riff = m:read_bytes(4)
    local size = m:read_dword()
    local aprg = m:read_bytes(4)
    if riff == "RIFF" and aprg == "APRG" then
        return true
    else
        return false
    end
end

function akp.case_sample_path(filename, default_path, sample_path)
    local path = engine:get_sample_path(filename, default_path, sample_path .. ".wav")
    if not io.exists(path) then
        path = engine:get_sample_path(filename, default_path, sample_path .. ".WAV")
    end
    return path
end

function akp.parse_kgrp(m, chunk_size, keymaps)
    -- Parses a kgrp chunk
end

function akp.parse_kloc(m, chunk_size, keymaps)
    -- Parses a kloc chunk
    m:skip(4)                                -- unsigned int res0           // 00AE - 00B1:   01,03,01,04
    local low_note = m:read_byte()           -- unsigned char low_note      // 00B2:    15    low note (21) 21 -> 127
    local high_note = m:read_byte()          -- unsigned char high_note     // 00B3:    7F    high note (127) 21 -> 127
    local semi_tune = m:read_byte_signed()   -- char semitone_tune          // 00B4:    00    Semitone Tune (0) -36 -> 36
    local fine_tune = m:read_byte_signed()   -- char fine_tune              // 00B5:    00    Fine Tune (0) -50 -> 50
    m:skip(1)                                -- unsigned char override_fx   // 00B6:    00    Overide FX (0) 0 = OFF, 1 = FX1, 2 = FX2, 3 = RV3, 4 = RV4
    m:skip(1)                                -- unsigned char fx_send_level // 00B7:    00    FX Send Level (0) 0 -> 100
    m:skip(1)                                -- char pitch_mod_1            // 00B8:    64    Pitch Mod 1 (100) -100 -> 100
    m:skip(1)                                -- char pitch_mod_2            // 00B9:    64    Pitch Mod 2 (100) -100 -> 100
    m:skip(1)                                -- char amp_mod                // 00BA:    00    Amp Mod (0) -100 -> 100
    m:skip(1)                                -- unsigned char zone_xfade    // 00BB:    00    Zone Xfade (0) 0 = OFF, 1 = ON
    m:skip(1)                                -- unsigned char mute_group    // 00BC:    00    Mute Group
    m:skip(1)                                -- unsigned char res1          // 00BD:    00    .

    -- import the data
    table.insert(keymaps, {
        low_note,
        high_note,
        semi_tune,
        fine_tune
    })
end

function akp.parse_zone(m, chunk_size, keymaps)
    -- Parses a zone chunk
    m:skip(1)                                            -- unsigned char res0          // 0126:    01    .
    local sample_name_len = math.min(m:read_byte(), 20)  -- unsigned char num_chars     // 0127:    nn    number of chars in Sample Name
    local sample_name = ""
    if sample_name_len > 0 then
        sample_name = m:read_bytes(sample_name_len)      -- char sample_name[20]        // 0128 - 013B:   Sample Name (pad with 00h) (if first character = 00h then no sample assigned)
    end
    m:skip(1)                                            -- char res1                   // 013C:    00    .
    m:skip(1)                                            -- char res2                   // 013D:    00    .
    m:skip(1)                                            -- char res3                   // 013E:    00    .
    m:skip(1)                                            -- char res4                   // 013F:    00    .
    m:skip(1)                                            -- char res5                   // 0140:    00    .
    m:skip(1)                                            -- char res6                   // 0141:    00    .
    m:skip(1)                                            -- char res7                   // 0142:    00    .
    m:skip(1)                                            -- char res8                   // 0143:    00    .
    m:skip(1)                                            -- char res9                   // 0144:    00    .
    m:skip(1)                                            -- char res10                  // 0145:    00    .
    m:skip(1)                                            -- char res11                  // 0146:    00    .
    m:skip(1)                                            -- char res12                  // 0147:    00    .
    local sample_min_vel = m:read_byte()                 -- unsigned char low_vel       // 0148:    00    Low Velocity (0) 0 -> 127
    local sample_max_vel = m:read_byte()                 -- unsigned char high_vel      // 0149:    7F    High Velocity (127) 0 -> 127
    local sample_fine_tune = m:read_byte_signed()        -- char fine_tune              // 014A:    00    Fine Tune (0) -50 -> 50
    local sample_semi_tune = m:read_byte_signed()        -- char semitone_tune          // 014B:    00    Semitone Tune (0) -36 -> 36
    m:skip(1)                                            -- char filter                 // 014C:    00    Filter (0) -100 -> 100
    local sample_pan = m:read_byte_signed()              -- char pan_balance            // 014D:    00    Pan/Balance (0) -50 -> 50 = L50 -> R50
    local sample_loop_mode = m:read_byte()               -- unsigned char loop_mode     // 014E:    04    Playback (4) 0 = NO LOOPING, 1 = ONE SHOT, 2 = LOOP IN REL, 3 = LOOP UNTIL REL, 4 = AS SAMPLE
    m:skip(1)                                            -- unsigned char output        // 014F:    00    Output (0)
    local sample_level = m:read_byte_signed()            -- char zone_level             // 0150:    00    Zone Level (0) -100 -> 100
    local sample_kbd_track = m:read_byte()               -- unsigned char kbd_track     // 0151:    01    Keyboard Track (1) 0 = OFF, 1 = ON
    m:skip(1)                                            -- unsigned char vel_start_msb // 0152:    00    Velocity->Start LSB (0)
    m:skip(1)                                            -- unsigned char vel_start_lsb // 0153:    00    Velocity->Start MSB (0) -9999 -> 9999

    -- import sample data
    table.insert(keymaps[#keymaps], sample_name)
    table.insert(keymaps[#keymaps], sample_semi_tune)
    table.insert(keymaps[#keymaps], sample_fine_tune)
    table.insert(keymaps[#keymaps], sample_min_vel)
    table.insert(keymaps[#keymaps], sample_max_vel)
    table.insert(keymaps[#keymaps], sample_pan)
    table.insert(keymaps[#keymaps], sample_level)
    table.insert(keymaps[#keymaps], sample_loop_mode)
    table.insert(keymaps[#keymaps], sample_kbd_track)
end

--==============================================================================
-- The actual importing function

local function akp_import_instrument(filename, program_index)
    logger.log("AKP: Started importing...")

    local default_path = ""
    local sample_path = ""
    local start_pos = 1
    local chunk_start = 1
    local working = true
    local keymaps = {}
    local sample_count = 0

    -- read file
    local m = utils.FileReader.create(filename)
    if m == nil then
        logger.log("AKP: Couldn't open Akai S5000/S6000 program file...")
        return 0
    end

    -- check the validity of the file
    if not akp.is_valid_file(m) then
        logger.log("AKP: " .. filename .. " is not a valid Akai S5000/S6000 program file...")
        return 0
    end

    -- extract the instrument path/name
    local inst_path = io.dirname(filename)
    local inst_name = io.filenoext(filename)

    -- parse keygroups
    while not m:finished() and (m:size() - m:tell()) > 8 do
        local chunk_id = m:read_bytes(4)
        local chunk_size = m:read_dword()
        local old_pos = m:tell()
        debug.log(" -- processing " .. chunk_id)
        if chunk_id == "kgrp" then
            akp.parse_kgrp(m, chunk_size, keymaps)
        elseif chunk_id == "kloc" then
            akp.parse_kloc(m, chunk_size, keymaps)
            m:seek(old_pos + chunk_size)
        elseif chunk_id == "zone" then
            akp.parse_zone(m, chunk_size, keymaps)
            m:seek(old_pos + chunk_size)
        else
            m:skip(chunk_size)
        end
    end

    m:close()

    -- check samples
    for i = 1, #keymaps do
        for z = 0, 3 do
            local samplename = keymaps[i][4 + z * 9 + 1]
            if samplename ~= "" then
                sample_path = akp.case_sample_path(filename, default_path, samplename)
                if io.exists(sample_path) then
                    sample_count = sample_count + 1
                end
            end
        end
    end

    if sample_count == 0 then
        logger.log("AKP: No samples found in Akai S5000/S6000 program...")
        return 0
    end

    -- iterate keymaps
    local prg = engine:get_program(0)
    prg:clear()
    prg:set_name(inst_name)

    for i = 1, (#keymaps) do
        local low_note = keymaps[i][1]
        local high_note = keymaps[i][2]
        local semi_tune = keymaps[i][3]
        local fine_tune = keymaps[i][4]

        for z = 0, 3 do
            local zone_index = 4 + z * 9
            local sample_name = keymaps[i][zone_index + 1]
            local sample_semi_tune = keymaps[i][zone_index + 2]
            local sample_fine_tune = keymaps[i][zone_index + 3]
            local sample_min_vel = keymaps[i][zone_index + 4]
            local sample_max_vel = keymaps[i][zone_index + 5]
            local sample_pan = keymaps[i][zone_index + 6]
            local sample_level = keymaps[i][zone_index + 7]
            local sample_loop_mode = keymaps[i][zone_index + 8]
            local sample_kbd_track = keymaps[i][zone_index + 9]

            local sample_file = akp.case_sample_path(filename, default_path, sample_name)
            if io.exists(sample_file) then
                local z = prg:add_zone()
                engine:load_sample(z, sample_file, true)

                z:set_name(engine:extract_name(sample_file))

                --z.loop_start (taken from file if present)
                --z.loop_end (taken from file if present)
                z.loop_mode = sample_loop_mode
                z.num_cues = 0 -- force reset cues

                z.lo_input_range.midi_key = low_note
                z.lo_input_range.midi_vel = sample_min_vel
                z.hi_input_range.midi_key = high_note
                z.hi_input_range.midi_vel = sample_max_vel

                z.midi_coarse_tune = semi_tune + sample_semi_tune
                z.midi_fine_tune = fine_tune + sample_fine_tune
                if not sample_kbd_track then
                    z.midi_keycents = 0
                end
                z.mp_gain = sample_level
                z.mp_pan = sample_pan
            end
        end
    end

    -- debug print
    --rprint(keymaps)

    logger.log("AKP: Finished importing 1 program/s...")
    return 1
end

--==============================================================================
-- Register Format with the Engine !

engine:register_format({
    wildcard="*.akp",
    import=akp_import_instrument,
    version=5,
})
