local utils = require("bliss.utils")

--==============================================================================
-- Helpers

local exs = {}

function exs.is_valid_file(m)
    m:seek(0)
    local chunk1 = m:read_bytes(4)
    local chunk2 = m:read_bytes(4)
    m:skip(8)
    local chunk_id = m:read_bytes(4)
    m:seek(124)

    if chunk1 == string.char(0x00, 0x01, 0x00, 0x00)
       and chunk_id == "SOBT" then
        -- big endian
        return string.char(0x00, 0x01, 0x00, 0x01),
               string.char(0x00, 0x01, 0x00, 0x02),
               string.char(0x00, 0x01, 0x00, 0x03),
               chunk2,
               false
    elseif chunk1 == string.char(0x01, 0x01, 0x00, 0x00)
           and chunk_id == "TBOS" then
        -- little endian
        return string.char(0x01, 0x01, 0x00, 0x01),
               string.char(0x01, 0x01, 0x00, 0x02),
               string.char(0x01, 0x01, 0x00, 0x03),
               chunk2,
               true
    end
    return "","","","",true
end

function exs.get_sample_path(inst_path, inst_name, sample_name)
    local search_subdirs = function(base_path, inst_name, sample_name)
        local sample_path = engine:get_sample_path(base_path, "", sample_name)
        if io.exists(sample_path) then
            return sample_path
        end

        local subdirs = { "samples", "Samples" }
        for i=1, #subdirs do
            sample_path = engine:find_sample_path(base_path, subdirs[i], sample_name)
            if io.exists(sample_path) then
                return sample_path
            end
        end

        local postfixes = { "", "samples", "-samples", " samples", "Samples", "-Samples", " Samples" }
        for i=1, #postfixes do
            sample_path = engine:find_sample_path(base_path, inst_name .. postfixes[i], sample_name)
            if io.exists(sample_path) then
                return sample_path
            end
        end

        return ""
    end

    local sample_path = ""

    sample_path = search_subdirs(inst_path, inst_name, sample_name)
    if sample_path ~= "" then
        return sample_path
    end

    sample_path = search_subdirs(io.dirname(inst_path), inst_name, sample_name)
    if sample_path ~= "" then
        return sample_path
    end

    sample_path = engine:find_sample_path(inst_path, "", sample_name)
    if io.exists(sample_path) then
        return sample_path
    end

    return engine:get_sample_path(inst_path, "", sample_name)
end

function exs.loop_mode_to_number(loop_mode)
    loop_mode = tonumber(loop_mode)
    if loop_mode == 0 then -- no loop
        return 0
    elseif loop_mode == 1 then -- normal
        return 1
    elseif loop_mode == 3 then -- crossfade
        return 5
    end
    return 1
end

function exs.parse_zone_chunk(m, zones, little_endian)
    -- parses a zone chunk
    local zone_name         = m:read_string(64)                 -- 0
    local zone_play_mode    = m:read_byte_signed()              -- 64
    local zone_tracking     = not bit.band(zone_play_mode, 1)
    local zone_oneshot      = bit.band(zone_play_mode, 2)
    local zone_rootnote     = m:read_byte()                     -- 65
    local zone_finetune     = m:read_byte_signed()              -- 66
    local zone_pan          = m:read_byte_signed() * 0.5        -- 67
    local zone_volume       = m:read_byte_signed()              -- 68
    local zone_volume_scale = m:read_byte_signed()              -- 69
    local zone_startnote    = m:read_byte()                     -- 70
    local zone_endnote      = m:read_byte()                     -- 71
    m:skip(1)
    local zone_minvel       = m:read_byte()                     -- 73
    local zone_maxvel       = m:read_byte()                     -- 74
    m:skip(1)
    local zone_frame_start  = m:read_dword_signed()             -- 76
    local zone_frame_end    = m:read_dword_signed()             -- 80
    local zone_loop_start   = m:read_dword_signed()             -- 84
    local zone_loop_end     = m:read_dword_signed()             -- 88
    local zone_crossfade    = m:read_dword_signed()             -- 92
    m:skip(1)
    local zone_loop_mode    = m:read_byte()                     -- 97
    m:skip(46)
    local zone_semitune     = m:read_byte()                     -- 144
    m:skip(1)
    local output_ref        = m:read_byte()                     -- 146
    m:skip(5)
    local group_ref         = m:read_dword_signed()             -- 152
    local sample_ref        = m:read_word()                     -- 156
    if not little_endian then
        m:skip(1)
        sample_ref        = m:read_word()
    end

    zone_rootnote = math.limit(0, 119, zone_rootnote)
    zone_startnote = math.limit(0, 119, zone_startnote)
    zone_endnote = math.limit(0, 119, zone_endnote)
    zone_pan = math.limit(-50, 50, zone_pan)

    -- import zone data
    local zone = {
        sample=sample_ref,
        group=group_ref,
        basenote=zone_rootnote,
        keytrack=100,
        lokey=zone_startnote,
        hikey=zone_endnote,
        lovel=zone_minvel,
        hivel=zone_maxvel,
        tune=zone_finetune,
        volume=zone_volume,
        pan=zone_pan,
        loop_start=zone_loop_start,
        loop_end=zone_loop_end,
        loop_mode=exs.loop_mode_to_number(zone_loop_mode),
        name=zone_name
    }

    table.insert(zones, zone)

    return zones
end

function exs.parse_sample_chunk(m, samples, little_endian, magic_sample)
    local samplename = m:read_string(64)
    local data_offset = m:read_dword()
    local sample_length = m:read_dword()
    local sample_rate = m:read_dword()
    local sample_bits = m:read_dword()
    local sample_channels = m:read_dword()
    local sample_channels_info = m:read_dword()
    m:skip(4)  -- Unknown ?
    local wave_id = m:read_bytes(4)
    local wave_rough = m:read_dword()
    m:skip(44) -- Unknown ?
    m:skip(256) -- Path ?
    m:skip(256) -- Name ?

    if wave_id == "EVAW"
       or wave_id == "WAVE"
       or wave_id == "FFIA"
       or wave_id == "AIFF"
    then
        table.insert(samples, samplename)
    else
        table.insert(samples, "invalid")
    end

    return samples
end

--==============================================================================
-- The actual importing function

local function exs_import_instrument(filename, program_index)
    logger.log("EXS: Started importing...")

    local inst_path = ""
    local inst_name = ""
    local sample_path = ""
    local zones     = {}
    local samples   = {}
    local magic_zone = ""
    local magic_group = ""
    local magic_sample = ""
    local chunk_format = ""
    local little_endian = true
    local file_format = 3

    -- load and process chunks
    local m = utils.FileReader.create(filename)
    if m == nil then
        logger.log("EXS: Couldn't open Logic EXS program file...")
        return 0
    end

    -- check the validity of the file
    magic_zone, magic_group, magic_sample, chunk_format, little_endian = exs.is_valid_file(m)
    if magic_zone == "" then
        logger.log("EXS: " .. filename .. " is not a valid Logic EXS program file...")
        return 0
    end

    -- file format
    local format_log = "Format is "

    if chunk_format == string.char(0x00, 0x00, 0x00, 0x28)
       or chunk_format == string.char(0x28, 0x00, 0x00, 0x00) then
        format_log = format_log .. "3"
    elseif chunk_format == string.char(0x00, 0x00, 0x00, 0x50)
           or chunk_format == string.char(0x50, 0x00, 0x00, 0x00) then
        format_log = format_log .. "5"
    end

    if little_endian then
        format_log = format_log .. " little endian"
    else
        format_log = format_log .. " big endian"
    end

    logger.log("ESX: " .. format_log)

    -- extract the instrument path/name
    local inst_path = io.dirname(filename)
    local inst_name = io.filenoext(filename)

    -- parse zone chunks
    local working = true
    while not m:finished() and working == true do
        local start_pos = m:tell()

        -- exit if we have no more data to read
        if m:size() < start_pos + 20 then
            break
        end

        -- read chunk
        chunk_id = m:read_bytes(4)

        -- determine chunk size
        if not little_endian then
            m:skip(3)
        end
        chunk_size = 20 + 64 + m:read_byte()

        -- offset on chunk name
        m:seek(start_pos + 16)
        chunk_name = m:read_bytes(4)

        -- parse chunks
        if chunk_id == magic_zone
           and (chunk_name == "TBOS" or chunk_name == "SOBT")
        then
            --logger.log("EXS: Parse zone")
            zones = exs.parse_zone_chunk(m, zones, little_endian)
            m:seek(start_pos + chunk_size)
        elseif chunk_id == magic_group
               and (chunk_name == "TBOS" or chunk_name == "SOBT")
        then
            --logger.log("EXS: Parse group")
            -- groups = exs.parse_group_chunk(m, groups, little_endian)
            m:seek(start_pos + chunk_size)
        elseif chunk_id == magic_sample
               and (chunk_name == "TBOS" or chunk_name == "SOBT")
        then
            --logger.log("EXS: Parse sample")
            samples = exs.parse_sample_chunk(m, samples, little_endian, magic_sample)
            m:seek(start_pos + chunk_size)
        else
            m:seek(start_pos + 2)
        end
    end

    m:close()

    -- check references
    if #zones == 0 then
        logger.log("EXS: Patch import found no zone references...")
        return 0
    end

    if #samples == 0 then
        logger.log("EXS: Patch import found no sample references...")
        return 0
    end

    -- build instrument
    local prg = engine:get_program(0)
    prg:clear()
    prg:set_name(inst_name)

    for i = 1, (#zones) do -- increase i for each zone
        local z = nil
        local zone = zones[i]
        local sample_name = samples[zone.sample + 1]
        local sample_path = exs.get_sample_path(filename, inst_name, sample_name)
        if io.exists(sample_path) then
            z = prg:add_zone()
            engine:load_sample(z, sample_path, true)
            z:set_name(engine:extract_name(sample_name))
            z:set_path(sample_path)
        else
            logger.log("EXS: Sample " .. sample_path .. " does not exists")
        end

        if z ~= nil then
            z.loop_start = zone.loop_start
            z.loop_end = zone.loop_end
            z.loop_mode = zone.loop_mode
            z.num_cues = 0 -- reset cues

            z.lo_input_range.midi_key = zone.lokey
            z.lo_input_range.midi_vel = zone.lovel
            z.hi_input_range.midi_key = zone.hikey
            z.hi_input_range.midi_vel = zone.hivel

            z.midi_root_key = zone.basenote
            z.midi_coarse_tune = 0
            z.midi_fine_tune = zone.tune
            z.midi_keycents = zone.keytrack

            z.mp_gain = zone.volume
            z.mp_pan = zone.pan
        end
    end

    -- debug print
    --rprint(samples)
    --rprint(zones)

    logger.log("EXS: Finished importing 1 program/s...")
    return 1
end

--==============================================================================
-- Register Format with the Engine !

engine:register_format({
    wildcard="*.exs",
    import=exs_import_instrument,
    version=5,
})
