local utils = require("bliss.utils")

--==============================================================================
-- Helpers

local sf2 = {}

function sf2.is_valid_file(m)
    m:seek(0) -- restart from beginning
    riff = m:read_bytes(4)
    size = m:read_dword()
    sfbk = m:read_bytes(4)
    if riff == "RIFF" and sfbk == "sfbk" then
        return true
    else
        return false
    end
end

function sf2.transform_unit(v)
    return v
end

function sf2.transform_signed(v)
    if v == 0x8000 then
        return -32786
    elseif v < 0x8000 then
        return v
    else
        return -32768 + (v - 0x8000)
    end
end

function sf2.transform_sample(v)
    return sf2.transform_unit(v)
end

function sf2.transform_32k_samples(v)
    return sf2.transform_signed(v)
end

function sf2.transform_cent_fs(v)
    return sf2.transform_signed(v)
end

function sf2.transform_cent(v)
    return 8.176 * math.pow(2.0, sf2.transform_signed(v) / 1200.0)
end

function sf2.transform_timecent(v)
    return math.pow(2.0, sf2.transform_signed(v) / 1200.0)
end

function sf2.transform_db(v)
    return (1000 - math.limit(0, 1000, sf2.transform_signed(v))) / 10.0
end

function sf2.transform_10th_percent(v)
    return sf2.transform_signed(v) / 10.0
end

function sf2.transform_key_range(v)
    return { bit.band(v, 0xFF), bit.band(bit.rshift(v, 8), 0xFF) }
end

function sf2.transform_vel_range(v)
    return { bit.band(v, 0xFF), bit.band(bit.rshift(v, 8), 0xFF) }
end

function sf2.transform_loop_flags(v)
    return math.limit(0, 3, v)
end

function sf2.generators_table(idx)
    local transform_table = {
        { "SF_startAddrs", sf2.transform_sample },                      -- 0 sample start address -4 (0to*0xffffff)
        { "SF_endAddrs", sf2.transform_sample },                        -- 1
        { "SF_startloopAddrs", sf2.transform_sample },                  -- 2 loop start address -4 (0 to * 0xffffff)
        { "SF_endloopAddrs", sf2.transform_sample },                    -- 3 loop end address -3 (0 to * 0xffffff)
        { "SF_startAddrsCoarseOffset", sf2.transform_32k_samples },     -- 4 high word of startAddrs
        { "SF_lfo1ToPitch", sf2.transform_cent_fs },                    -- 5 main fm: lfo1-> pitch
        { "SF_lfo2ToPitch", sf2.transform_cent_fs },                    -- 6 aux fm:  lfo2-> pitch
        { "SF_env1ToPitch", sf2.transform_cent_fs },                    -- 7 pitch env: env1(aux)-> pitch
        { "SF_initialFilterFc", sf2.transform_cent },                   -- 8 initial filter cutoff
        { "SF_initialFilterQ", sf2.transform_db },                      -- 9 filter Q
        { "SF_lfo1ToFilterFc", sf2.transform_cent_fs },                 -- 10 filter modulation: lfo1->filter*cutoff
        { "SF_env1ToFilterFc", sf2.transform_cent_fs },                 -- 11 filter env: env1(aux)->filter * cutoff
        { "SF_endAddrsCoarseOffset", sf2.transform_32k_samples },       -- 12 high word of endAddrs
        { "SF_lfo1ToVolume", sf2.transform_db },                        -- 13 tremolo: lfo1-> volume
        { "SF_unused1", sf2.transform_unit },                           -- 14 unused
        { "SF_chorusEffectsSend", sf2.transform_10th_percent },         -- 15 chorus
        { "SF_reverbEffectsSend", sf2.transform_10th_percent },         -- 16 reverb
        { "SF_pan", sf2.transform_10th_percent },                       -- 17 pan
        { "SF_unused2", sf2.transform_unit },                           -- 18 unused
        { "SF_unused3", sf2.transform_unit },                           -- 19 unused
        { "SF_unused4", sf2.transform_unit },                           -- 20 unused
        { "SF_delayLfo1", sf2.transform_timecent },                     -- 21 delay 0x8000-n*(725us)
        { "SF_freqLfo1", sf2.transform_cent },                          -- 22 frequency
        { "SF_delayLfo2", sf2.transform_timecent },                     -- 23 delay 0x8000-n*(725us)
        { "SF_freqLfo2", sf2.transform_cent },                          -- 24 frequency
        { "SF_delayEnv1", sf2.transform_timecent },                     -- 25 delay 0x8000 - n(725us)
        { "SF_attackEnv1", sf2.transform_timecent },                    -- 26 attack
        { "SF_holdEnv1", sf2.transform_timecent },                      -- 27 hold
        { "SF_decayEnv1", sf2.transform_timecent },                     -- 28 decay
        { "SF_sustainEnv1", sf2.transform_db },                         -- 29 sustain
        { "SF_releaseEnv1", sf2.transform_timecent },                   -- 30 release
        { "SF_keynumToEnv1Hold", sf2.transform_10th_percent },          -- 31
        { "SF_keynumToEnv1Decay", sf2.transform_10th_percent },         -- 32
        { "SF_delayEnv2", sf2.transform_timecent },                     -- 33 delay 0x8000 - n(725us)
        { "SF_attackEnv2", sf2.transform_timecent },                    -- 34 attack
        { "SF_holdEnv2", sf2.transform_timecent },                      -- 35 hold
        { "SF_decayEnv2", sf2.transform_timecent },                     -- 36 decay
        { "SF_sustainEnv2", sf2.transform_db },                         -- 37 sustain
        { "SF_releaseEnv2", sf2.transform_timecent },                   -- 38 release
        { "SF_keynumToEnv2Hold", sf2.transform_unit },                  -- 39
        { "SF_keynumToEnv2Decay", sf2.transform_unit },                 -- 40
        { "SF_instrument", sf2.transform_unit },                        -- 41
        { "SF_reserved1", sf2.transform_unit },                         -- 42 reserved
        { "SF_keyRange", sf2.transform_key_range },                     -- 43
        { "SF_velRange", sf2.transform_vel_range },                     -- 44
        { "SF_startloopAddrsCoarseOffset", sf2.transform_32k_samples }, -- 45 high word of startloopAddrs
        { "SF_keynum", sf2.transform_key_range },                       -- 46
        { "SF_velocity", sf2.transform_vel_range },                     -- 47
        { "SF_initAtten", sf2.transform_db },                           -- 48
        { "SF_reserved2", sf2.transform_unit },                         -- 49 reserved
        { "SF_endloopAddrsCoarseOffset", sf2.transform_32k_samples },   -- 50 high word of endloopAddrs
        { "SF_coarseTune", sf2.transform_signed },                      -- 51
        { "SF_fineTune", sf2.transform_cent },                          -- 52
        { "SF_sampleId", sf2.transform_unit },                          -- 53
        { "SF_sampleModes", sf2.transform_loop_flags },                 -- 54
        { "SF_reserved3", sf2.transform_unit },                         -- 55 reserved
        { "SF_scaleTuning", sf2.transform_signed },                     -- 56
        { "SF_exclusiveClass", sf2.transform_unit },                    -- 57
        { "SF_overridingRootKey", sf2.transform_unit },                 -- 58
        { "SF_unused5", sf2.transform_unit },                           -- 59 unused
        { "SF_endOper", sf2.transform_unit },                           -- 60
    }
    -- return local tab
    local tab = transform_table[idx + 1]
    return { tab[1], tab[2] }
end


function sf2.process_list(m, size, sf)
    chunk_id = m:read_bytes(4)
    size = size - 4

    if chunk_id == "INFO" then
        return sf2.process_info(m, size, sf)
    elseif chunk_id == "sdta" then
        return sf2.process_sdta(m, size, sf)
    elseif chunk_id == "pdta" then
        return sf2.process_pdta(m, size, sf)
    else
        m:skip(size)
    end
    return 0
end

function sf2.process_info(m, size, sf)
    while size > 0 do
        chunk_id = m:read_bytes(4)
        chunk_size = m:read_dword()
        size = size - 8

        if m:finished() then return -1 end

        if chunk_id == "ifil" then
            sf.version = m:read_word()
            sf.minorversion = m:read_word()
        elseif chunk_id == "INAM" then
            sf.name = m:read_string(chunk_size)
        else
            m:skip(chunk_size)
        end

        size = size - chunk_size
    end
    return 0
end

function sf2.process_sdta(m, size, sf)
    while size > 0 do
        chunk_id = m:read_bytes(4)
        chunk_size = m:read_dword()
        size = size - 8

        if m:finished() then return -1 end

        if chunk_id == "snam" then
            if sf.version > 1 then
                logger.log("SF2: Version 2 has obsolete format...")
                m:skip(chunk_size)
            else
                nsamples = math.round(chunk_size / 20)
                if sf.samples == nil then
                    sf.samples = {}
                    for i=1, nsamples do
                        sample_name = m:read_string(20)
                        table.insert(sf.samples, { name=sample_name })
                    end
                else
                    m:skip(size)
                end
            end
        elseif chunk_id == "smpl" then
            sf.samplepos = m:tell()
            sf.samplesize = chunk_size / 2;
            sf.sampledata = m:read_bytes(chunk_size)
        elseif chunk_id == "sm24" then
            logger.log("SF2: Unsupported sm24 sample format...")
            m:skip(chunk_size)
        else
            m:skip(chunk_size)
        end

        size = size - chunk_size
    end
    return 0
end

function sf2.process_pdta(m, size, sf)
    while size > 0 do
        chunk_id = m:read_bytes(4)
        chunk_size = m:read_dword()
        size = size - 8

        if m:finished() then return -1 end

        if chunk_id == "phdr" then
            sf2.process_preset_headers(m, size, sf)
        elseif chunk_id == "pbag" then
            sf2.process_bag(m, chunk_size, sf.prbags)
        elseif chunk_id == "pgen" then
            sf2.process_gen(m, chunk_size, sf.prbags)
        elseif chunk_id == "inst" then
            sf2.process_inst(m, chunk_size, sf)
        elseif chunk_id == "ibag" then
            sf2.process_bag(m, chunk_size, sf.inbags)
        elseif chunk_id == "igen" then
            sf2.process_gen(m, chunk_size, sf.inbags)
        elseif chunk_id == "shdr" then
            sf2.process_samples(m, chunk_size, sf)
        else
            m:skip(chunk_size)
        end

        size = size - chunk_size
    end
    return 0;
end

function sf2.process_preset_headers(m, size, sf)
    npresets = math.round(size / 38)
    sf.presets = {}
    for i=1, npresets do
        name = m:read_string(20)
        preset = m:read_word()
        bank = m:read_word()
        bagNdx = m:read_word()
        lib = m:read_dword() -- unused
        genre = m:read_dword() -- unused
        morph = m:read_dword() -- unused
        table.insert(sf.presets, {
            name=name,
            preset=preset,
            bank=bank,
            bagNdx=bagNdx,
            layers={},
        })
        if name == "EOP" then break end
    end
end

function sf2.process_inst(m, size, sf)
    ninsts = math.round(size / 22);
    sf.inst = {}
    for i=1, ninsts do
        name = m:read_string(20)
        bagNdx = m:read_word()
        table.insert(sf.inst, {
            name=name,
            bagNdx=bagNdx,
            layers={},
        })
        if name == "EOI" then break end
    end
end

function sf2.process_samples(m, size, sf)
    if sf.version >= 2 then
        -- SF2: include sample name and other infos
        nsamples = math.round(size / 46)
        sf.samples = {}
    else
        -- SBK: sample name may be read already
        nsamples = math.round(size / 16)
        if sf.samples == nil or #sf.samples == 0 then
            sf.samples = {}
        end
    end

    in_rom = 1
    for i=1, nsamples do
        local smpl = {}
        if sf.version >= 2 then -- SF2 only
            smpl.name = m:read_string(20)
        end

        smpl.startsample = m:read_dword()
        smpl.endsample = m:read_dword()
        smpl.startloop = m:read_dword()
        smpl.endloop = m:read_dword()
        if sf.version >= 2 then -- SF2 only
            smpl.samplerate = m:read_dword()
            smpl.originalPitch = m:read_byte()
            smpl.pitchCorrection = m:read_byte()
            smpl.samplelink = m:read_word()
            smpl.sampletype = m:read_word()
        else
            smpl.samplerate = 44100
            smpl.originalPitch = 60
            smpl.pitchCorrection = 0
            smpl.samplelink = 0
            if smpl.startsample == 0 then
                in_rom = 0
            end
            if in_rom then
                smpl.sampletype = 0x8001
            else
                smpl.sampletype = 1
            end
        end

        table.insert(sf.samples, smpl)

        if name == "EOS" then break end
    end
end

function sf2.process_gen(m, size, bags)
    size = math.round(size / 4)
    bags.gen = {}
    for i=1, size do
        operator = m:read_word()
        amount = m:read_word()
        if operator >= 0 and operator < 60 then
            -- calculate values
            gen = sf2.generators_table(operator)
            table.insert(bags.gen, {
                operator=operator,
                operator_type=gen[1],
                amount=amount,
                real_amount=gen[2](amount)
            })
        else
            logger.log("SF2: Unsupported " .. operator .. " operator type...")
        end
    end
end

function sf2.process_bag(m, size, bags)
    size = math.round(size / 4)
    bags.bag = {}
    for i=1, size do
        table.insert(bags.bag, {
            value=m:read_word(),
            modul=m:read_word() -- unused
        })
    end
end

function sf2.generate_layers(hdr, next, bags)
    nlayers = next.bagNdx - hdr.bagNdx;
    if nlayers == 0 then
        return
    end
    if nlayers < 0 then
        logger.log("SF2: Illegal layer number " .. nlayers .. "...")
        return
    end

    for i=hdr.bagNdx, next.bagNdx - 1 do
        table.insert(hdr.layers, { gen={} })

        bagIdx = i + 1
        genNdx = bags.bag[bagIdx].value
        nlists = bags.bag[bagIdx + 1].value - genNdx;
        if nlists < 0 then
            logger.log("SF2: Illegal list number " .. nlists .. "...")
        end

        layp = hdr.layers[#hdr.layers]
        for t=1, nlists do
            table.insert(layp.gen, bags.gen[genNdx + t])
        end
    end
end

function sf2.convert_layers(sf)
    if (#sf.prbags.bag == 0 or #sf.prbags.gen == 0) or
       (#sf.inbags.bag == 0 or #sf.inbags.gen == 0)
    then
        logger.log("SF2: Illegal bags / gens in file...")
        return false
    end

    -- generate preset layers
    for i=1, #sf.presets - 1 do
        sf2.generate_layers(sf.presets[i], sf.presets[i + 1], sf.prbags)
    end

    -- generate instrument layers
    for i=1, #sf.inst - 1 do
        sf2.generate_layers(sf.inst[i], sf.inst[i + 1], sf.inbags)
    end

    return true
end

function sf2.correct_samples(sf)
    prev_end = 0
    for i=1, #sf.samples do
        sp = sf.samples[i]

        -- correct sbk fileformat
        if sf.version == 1 then
            sp.startloop = sp.startloop + 1;
            sp.endloop = sp.endloop + 2;
        end

        -- calculate sample data size
        if bit.band(sp.sampletype, 0x8000) ~= 0 then
            sp.size = 0
        elseif sp.startsample < prev_end and sp.startsample ~= 0 then
            sp.size = 0
        else
            sp.size = -1
            if i ~= #sf.samples then
                sp1 = sf.samples[i + 1]
                sp.size = sp1.startsample - sp.startsample
            end
            if sp.size < 0 then
                sp.size = sp.endsample - sp.startsample + 48
            end
        end
        prev_end = sp.endsample

        -- calculate short-shot loop size
        if i == #sf.samples then
            sp.loopshot = 48
        else
            sp1 = sf.samples[i + 1]
            sp.loopshot = sp1.startsample - sp.endsample;
            if sp.loopshot < 0 or sp.loopshot > 48 then
                sp.loopshot = 48
            end
       end
    end
end

--==============================================================================
-- The actual importing function

local function sf2_import_instrument(filename, program_index)
    logger.log("SF2: Started importing...")

    -- load and process chunks
    local m = utils.FileReader.create(filename)
    if m == nil then
        logger.log("SF2: Couldn't open Soundfont2 program file...")
        return 0
    end

    -- check the validity of the file
    if not sf2.is_valid_file(m) then
        logger.log("SF2: " .. filename .. " is not a valid Soundfont2 program file...")
        return 0
    end

    -- extract the instrument path/name
    local inst_path = io.dirname(filename)
    local inst_name = io.filenoext(filename)

    -- parse file
    local sf = {
        prbags = {},
        inbags = {},
    }

    local retval = 0
    while not m:finished() do
        local chunk_id = m:read_bytes(4)
        local chunk_size = m:read_dword()

        if chunk_id == "LIST" then
            -- process list
            retval = sf2.process_list(m, chunk_size, sf)
        else
            -- skip
            m:skip(chunk_size)
        end
    end

    m:close()

    sf2.correct_samples(sf)
    sf2.convert_layers(sf)

    -- reset temporary
    sf.prbags = {}
    sf.inbags = {}

    -- convert to bliss zones
    local program_count = 0
    for i=1, #sf.inst - 1 do
        local inst = sf.inst[i]
        if inst.name == nil then
            inst.name = inst_name
        end

        if program_count >= (engine:num_programs() - program_index) then
            break
        end

        local prg = engine:get_program(program_count)
        prg:clear()
        prg:set_name(inst.name)

        for l=1, #inst.layers do
            local layer = inst.layers[l]

            local sample_id = nil
            local low_note = 0
            local high_note = 119
            local root_key = -1
            local coarse_tune = 0
            local fine_tune = 0
            local scale_tuning = 100
            local sample_min_vel = 0
            local sample_max_vel = 127
            local sample_loop_mode = 0
            local sample_pan = 0
            local chorus_fx = 0
            local reverb_fx = 0
            local startAddrsOffset = 0
            local endAddrsOffset = 0
            local startAddrsCoarseOffset = 0
            local endAddrsCoarseOffset = 0
            local startloopAddrsOffset = 0
            local endloopAddrsOffset = 0
            local startloopAddrsCoarseOffset = 0
            local endloopAddrsCoarseOffset = 0
            local init_atten = 0
            local ampenv_attack = 0
            local ampenv_decay = 0
            local ampenv_sustain = 100
            local ampenv_release = 0
            local modenv_attack = 0
            local modenv_decay = 0
            local modenv_sustain = 100
            local modenv_release = 0

            for g=1, #layer.gen do
                local gen = layer.gen[g]
                --if type(gen.real_amount) == "table" then
	            --    logger.log("SF2: Program " .. program_count .. " gen " .. gen.operator_type .. " = {}...")
	            --else
	            --    logger.log("SF2: Program " .. program_count .. " gen " .. gen.operator_type .. " = " .. gen.real_amount .. "...")
	            --end

                if gen.operator_type == "SF_sampleId" then
                    sample_id = gen.real_amount
                elseif gen.operator_type == "SF_startAddrsOffset" then
                    startAddrsOffset = gen.real_amount
                elseif gen.operator_type == "SF_endAddrsOffset" then
                    endAddrsOffset = gen.real_amount
                elseif gen.operator_type == "SF_startAddrsCoarseOffset" then
                    startAddrsCoarseOffset = gen.real_amount
                elseif gen.operator_type == "SF_endAddrsCoarseOffset" then
                    endAddrsCoarseOffset = gen.real_amount
                elseif gen.operator_type == "SF_startloopAddrsOffset" then
                    startloopAddrsOffset = gen.real_amount
                elseif gen.operator_type == "SF_endloopAddrsOffset" then
                    endloopAddrsOffset = gen.real_amount
                elseif gen.operator_type == "SF_startloopAddrsCoarseOffset" then
                    startloopAddrsCoarseOffset = gen.real_amount
                elseif gen.operator_type == "SF_endloopAddrsCoarseOffset" then
                    endloopAddrsCoarseOffset = gen.real_amount
                elseif gen.operator_type == "SF_keyRange" then
                    low_note = gen.real_amount[1]
                    high_note = gen.real_amount[2]
                elseif gen.operator_type == "SF_velRange" then
                    sample_min_vel = gen.real_amount[1]
                    sample_max_vel = gen.real_amount[2]
                elseif gen.operator_type == "SF_sampleModes" then
                    sample_loop_mode = math.limit(0, 1, gen.real_amount)
                elseif gen.operator_type == "SF_overridingRootKey" then
                    root_key = gen.real_amount
                elseif gen.operator_type == "SF_coarseTune" then
                    coarse_tune = gen.real_amount
                elseif gen.operator_type == "SF_fineTune" then
                    fine_tune = gen.real_amount
                elseif gen.operator_type == "SF_scaleTuning" then
                    scale_tuning = gen.real_amount
                elseif gen.operator_type == "SF_pan" then
                    sample_pan = gen.real_amount
                elseif gen.operator_type == "SF_initAtten" then
                    init_atten = gen.real_amount
                elseif gen.operator_type == "SF_chorusEffectsSend" then
                    chorus_fx = gen.real_amount
                elseif gen.operator_type == "SF_reverbEffectsSend" then
                    reverb_fx = gen.real_amount
                elseif gen.operator_type == "SF_attackEnv1" then
                	modenv_attack = gen.real_amount
                elseif gen.operator_type == "SF_decayEnv1" then
                	modenv_decay = gen.real_amount
                elseif gen.operator_type == "SF_sustainEnv1" then
                	modenv_release = gen.real_amount
                elseif gen.operator_type == "SF_releaseEnv1" then
                	ampenv_release = gen.real_amount
                elseif gen.operator_type == "SF_attackEnv2" then
                	ampenv_attack = gen.real_amount
                elseif gen.operator_type == "SF_decayEnv2" then
                	ampenv_decay = gen.real_amount
                elseif gen.operator_type == "SF_sustainEnv2" then
                	ampenv_sustain = gen.real_amount
                elseif gen.operator_type == "SF_releaseEnv2" then
                	modenv_sustain = gen.real_amount
                elseif gen.operator_type == "SF_freqLfo1" then
                elseif gen.operator_type == "SF_delayLfo1" then
                elseif gen.operator_type == "SF_lfo1ToVolume" then
                else
                    --logger.log("SF2: Program " .. program_count .. " unknown gen " .. gen.operator_type .. "...")
                end
            end

            if sample_id ~= nil then
                local smpl = sf.samples[sample_id + 1]

                local z = prg:add_zone()
                z:set_name(smpl.name)

                samplestart = smpl.startsample + startAddrsOffset + startAddrsCoarseOffset
                endsample = smpl.endsample + endAddrsOffset + endAddrsCoarseOffset
                samplesize = endsample - samplestart

                engine:load_sample_sint16(z, 1, samplesize, samplestart, sf.sampledata)

                z.loop_start = (smpl.startloop - smpl.startsample) + startloopAddrsOffset + startloopAddrsCoarseOffset
                z.loop_end = (smpl.endloop - smpl.startsample) + endloopAddrsOffset + endloopAddrsCoarseOffset
                z.loop_mode = sample_loop_mode
                z.num_cues = 0 -- force reset cues

                z.lo_input_range.midi_key = low_note
                z.lo_input_range.midi_vel = sample_min_vel
                z.hi_input_range.midi_key = high_note
                z.hi_input_range.midi_vel = sample_max_vel

                if root_key < 0 then
                	root_key = smpl.originalPitch
                end
                z.midi_root_key = root_key
                z.midi_keycents = scale_tuning
                if coarse_tune ~= 0 then
                    z.midi_coarse_tune = coarse_tune
                end
                if fine_tune ~= 0 then
                    z.midi_fine_tune = fine_tune
                end
               --z.mp_gain = sample_level
                z.mp_pan = sample_pan

                z:set_ampenv_attack(ampenv_attack)
                z:set_ampenv_decay(ampenv_decay)
                z:set_ampenv_sustain(ampenv_sustain)
                z:set_ampenv_release(ampenv_release)
                z:set_modenv_attack(modenv_attack)
                z:set_modenv_decay(modenv_decay)
                z:set_modenv_sustain(modenv_sustain)
                z:set_modenv_release(modenv_release)
            else
                --logger.log("SF2: Program " .. program_count .. " cannot find SF_sampleId value in gen...")
            end
        end

        program_count = program_count + 1
    end

    -- debug print
    --rprint(sf)

    logger.log("SF2: Finished importing " .. program_count .. " program/s...")
    return program_count
end

--==============================================================================
-- Register Format with the Engine !

engine:register_format({
    wildcard="*.sf2;*.sbk",
    import=sf2_import_instrument,
    version=5,
})
