local utils = require("bliss.utils")

--==============================================================================
-- Helpers

local sfz = {}

function sfz.note_to_number(name_string)
    local n = 0
    if tonumber(name_string) ~= nil then
        n = tonumber(name_string)
    else
        local notes = "c d ef g a b"
        local note = notes:find(name_string:sub(1, 1):lower()) - 1
        local o = 0

        if tonumber(name_string:sub(2, 2)) ~= nil then
            -- white note
            o = tonumber(name_string:sub(2, 2))
        else
            -- black note
            if tonumber(name_string:sub(3, 3)) ~= nil then
                o = tonumber(name_string:sub(3, 3))
                if name_string:sub(2, 2):lower() == "b" then
                    note = note - 1
                elseif name_string:sub(2,2) == "#" then
                    note = note + 1
                end
            else
                return nil
            end
        end

        n = (o * 12) + note
    end
    return math.limit(0, 119, n)
end

function sfz.loop_mode_to_number(loop_mode)
    if tonumber(loop_mode) ~= nil then
        return math.limit(0, 4, tonumber(loop_mode))
    else
        if loop_mode == "no_loop" then
            return 0
        elseif loop_mode == "loop_continuous" then
            return 1
        elseif loop_mode == "loop_bidirectional" then
            return 2
        elseif loop_mode == "loop_backward" then
            return 3
        elseif loop_mode == "loop_sustain" then
            return 4
        end
        return 0
    end
end

function sfz.trigger_to_number(trigger_mode)
    if trigger_mode == "release" then
        return 1
    else
        return 0
    end
end

--==============================================================================
-- The actual importing function

local function sfz_import_instrument(filename, program_index)
    logger.log("SFZ: Started importing...")

    local lines = {}
    local keymaps = {}
    local default_path = ""

    local sample = ""
    local offset = -1
    local endz = -1
    local lokey = 0             -- 0..127
    local hikey = 119           -- 0..119
    local lovel = 0             -- 0..127
    local hivel = 127           -- 0..127
    local basenote = 60         -- 0..119
    local keytrack = 100        -- 0..1200
    local tune = 0              -- pitch offset in semitones
    local transpose = 0
    local volume = 0            -- in decibels (max 0db)
    local pan = 0               -- -63..0..63
    local loop_mode = -1
    local loop_start = 0
    local loop_end = 0
    local lobend = 0
    local hibend = 0
    local lochanaft = 0
    local hichanaft = 0
    local lopolyaft = 0
    local hipolyaft = 0
    local trigger = 0
    local group = -1
    local offby = -1
    local locc = {}
    local hicc = {}

    -- values
    local fill_values = function()
        table.insert(keymaps, {
            sample=sample,
            offset=offset,
            endz=endz,
            basenote=basenote,
            keytrack=keytrack,
            lokey=lokey,
            hikey=hikey,
            lovel=lovel,
            hivel=hivel,
            tune=tune,
            transpose=transpose,
            volume=volume,
            pan=pan,
            loop_mode=loop_mode,
            loop_start=loop_start,
            loop_end=loop_end,
            lobend=lobend,
            hibend=hibend,
            lochanaft=lochanaft,
            hichanaft=hichanaft,
            lopolyaft=lopolyaft,
            hipolyaft=hipolyaft,
            trigger=trigger,
            group=group,
            offby=offby,
            locc=locc,
            hicc=hicc,
        })
    end

    local reset_values = function()
        sample = ""
        offset = -1
        endz = -1
        lokey = 0
        hikey = 119
        lovel = 0
        hivel = 127
        basenote = 60
        keytrack = 100
        tune = 0
        transpose = 0
        volume = 0
        pan = 0
        loop_mode = -1
        loop_start = 0
        loop_end = 0
        lobend = 0
        hibend = 0
        lochanaft = 0
        hichanaft = 0
        lopolyaft = 0
        hipolyaft = 0
        trigger = 0
        group = -1
        offby = -1
        locc = {}
        hicc = {}
    end

    -- extract the instrument path/name
    local inst_path = io.dirname(filename)
    local inst_name = io.filenoext(filename)

    -- preparse file
    for line in io.lines(filename) do
        -- strip comments to end of line
        if line:find("//") then
          line = line:sub(1, line:find("//") - 1)
        end
        if line:find("/%*") then
            line = line:sub(1, line:find("/%*") - 1)
        end
        if line:find("%*/") then
            line = line:sub(1, line:find("%*/") - 1)
        end

        -- remove redundant CR on unix systems
        local cr_replaced = 0
        line, cr_replaced = line:gsub(string.char(13), "")

        -- convert "\" to "/" and trim
        line = line:gsub("\\", "/")
        line = line:gsub("^%s*(.-)%s*$", "%1")

        -- split and add
        for index, l in ipairs(string.split(line, " ")) do
            if      string.find(l, "%=") == nil
                and string.find(l, "^<[%w_]*>$") == nil
                and #lines > 0
            then
               lines[#lines] = lines[#lines] .. " " .. l
            else
               if l ~= "" then
                 table.insert(lines, l)
              end
            end
        end
    end

    -- loop over regions
    local first_region = true
    local inside_group = false
    for t=1, #lines do
        local pair = string.split(lines[t], "=")
        local key = pair[1]
        local value = pair[2]

        if key == "<control>" then
        elseif key == "<global>" then
        elseif key == "<master>" then
        elseif key == "<group>" then
            if sample ~= "" and first_region ~= true then
                fill_values()
                reset_values()
            end
            inside_group = true
            first_region = true

        elseif key == "<region>" then
            if sample ~= "" and first_region ~= true then
                fill_values()
                if not inside_group then
                    reset_values()
                end
            end
            first_region = false

        elseif key == "default_path" then
            if value ~= nil then
                default_path = value:gsub("\\", "/")
            end

        elseif key == "sample" then
            sample = value

        elseif key == "offset" then
            offset = tonumber(value)

        elseif key == "end" then
            endz = tonumber(value)

        elseif key == "lokey" then
            lokey = sfz.note_to_number(value)

        elseif key == "hikey" then
            hikey = sfz.note_to_number(value)

        elseif key == "lovel" then
            lovel = tonumber(value)

        elseif key == "hivel" then
            hivel = tonumber(value)

        elseif key == "tune" then
            tune = tonumber(value)

        elseif key == "transpose" then
            transpose = tonumber(value)

        elseif key == "volume" then
            volume = tonumber(value)

        elseif key == "pan" then
            pan = tonumber(value)

        elseif key == "loop_mode" then
            loop_mode = sfz.loop_mode_to_number(value)

        elseif key == "loop_start" then
            loop_start = tonumber(value)

        elseif key == "loop_end" then
            loop_end = tonumber(value)

        elseif key == "key" then
            lokey = sfz.note_to_number(value)
            hikey = lokey
            basenote = lokey

        elseif key == "pitch_keytrack" then
            keytrack = tonumber(value)

        elseif key == "pitch_keycenter" then
            basenote = sfz.note_to_number(value)

        elseif key == "lobend" then
            lobend = tonumber(value)

        elseif key == "hibend" then
            hibend = tonumber(value)

        elseif key == "lochanaft" then
            lochanaft = tonumber(value)

        elseif key == "hichanaft" then
            hichanaft = tonumber(value)

        elseif key == "lopolyaft" then
            lopolyaft = tonumber(value)

        elseif key == "hipolyaft" then
            hipolyaft = tonumber(value)

        elseif key == "trigger" then
            trigger = sfz.trigger_to_number(value)

        elseif key == "group" then
            group = tonumber(value)

        elseif key == "offby" then
            offby = tonumber(value)

        elseif key:find("locc(%d*)") then
            _, _, num = key:find("locc(%d*)")
            locc[num] = tonumber(value)

        elseif key:find("hicc(%d*)") then
            _, _, num = key:find("hicc(%d*)")
            hicc[num] = tonumber(value)

        else
            --logger.log("---- Unknown " .. key)
        end
    end

    -- last region
    fill_values()

    -- make safety checks !
    local found_maps = 0
    for t=1, #keymaps do
        sample_file = engine:get_sample_path(filename, default_path, keymaps[t].sample)
        if io.exists(sample_file) then
            found_maps = found_maps + 1
        end
    end

    if #keymaps == 0 or found_maps == 0 then
        logger.log("SFZ: Patch import found no sample references...")
        return 0
    end

    -- build instrument
    local prg = engine:get_program(0)
    prg:clear()
    prg:set_name(inst_name)

    for t=1, #keymaps do
        local km = keymaps[t]

        local sample_file = engine:get_sample_path(filename, default_path, km.sample)
        if io.exists(sample_file) then
            local z = prg:add_zone()

            if km.offset >= 0 and km.endz > 0 then
                engine:load_sample_offset(z, sample_file, true, km.offset, km.endz)
            else
                engine:load_sample(z, sample_file, true)
            end

            z:set_name(engine:extract_name(sample_file))

            if km.loop_mode ~= -1 then
                z.loop_start = km.loop_start - km.offset
                z.loop_end = km.loop_end  - km.offset
                z.loop_mode = km.loop_mode
            end
            z.num_cues = 0 -- force reset cues

            z.lo_input_range.midi_key = km.lokey
            z.lo_input_range.midi_vel = km.lovel
            z.hi_input_range.midi_key = km.hikey
            z.hi_input_range.midi_vel = km.hivel

            z.midi_root_key = km.basenote
            z.midi_coarse_tune = km.transpose
            z.midi_fine_tune = km.tune
            z.midi_keycents = km.keytrack

            z.mp_gain = km.volume
            z.mp_pan = km.pan * 0.5
        end
    end

    -- debug print
    --rprint(keymaps)

    logger.log("SFZ: Finished importing 1 program/s...")
    return 1
end

--==============================================================================
-- Register Format with the Engine !

engine:register_format({
    wildcard="*.sfz",
    import=sfz_import_instrument,
    version=5,
})
