local utils = require("bliss.utils")

--==============================================================================
-- Helpers

local xrni = {}

function xrni.table_contains(tab, key)
    return tab[key] ~= nil
end

function xrni.transform_none(value)
    return value
end

function xrni.transform_boolean(value)
    if value == "true" then
        return true
    end
    return false
end

function xrni.transform_note(value)
    return math.limit(0, 119, tonumber(value))
end

function xrni.transform_velocity(value)
    return math.limit(0, 127, tonumber(value))
end

function xrni.transform_volume(value)
    return dsp.lin2db(value)
end

function xrni.transform_pan(value)
    return 64 * math.limit(-0.5, 0.5, tonumber(value) - 0.5)
end

function xrni.transform_transpose(value)
    return tonumber(value)
end

function xrni.transform_fine_tune(value)
    return tonumber(value)
end

function xrni.transform_beat_sync_lines(value)
    return tonumber(value)
end

function xrni.transform_beat_sync_active(value)
    return xrni.transform_boolean(value)
end

function xrni.transform_loop_mode(value)
    if value == "Off" then
        return 0
    elseif value == "Forward" then
        return 1
    elseif value == "PingPong" then
        return 2
    elseif value == "Backward" then
        return 3
    end
    return 0
end

function xrni.transform_loop_start(value)
    return tonumber(value)
end

function xrni.transform_loop_end(value)
    return tonumber(value)
end

--==============================================================================
-- The actual importing function

local function xrni_import_instrument(filename, program_index)
    logger.log("XRNI: Started importing...")

    -- extract the instrument path/name
    local inst_path = io.dirname(filename)
    local inst_name = io.filenoext(filename)

    -- print the filenames of the files inside the zip
    local zfile = zip.ZipFile(filename)
    local zipped_files = zfile:files()

    local sample_files = {}
    for z = 1, #zipped_files do
        if zipped_files[z]:find("^SampleData") ~= nil then
            table.insert(sample_files, zipped_files[z])
        end
    end

    if #sample_files == 0 then
        logger.log("XRNI: No sample files found in instrument...")
        return 0
    end

    local myxml = zfile:unpack_file("Instrument.xml")
    if myxml ~= nil then
        -- Prepare parser
        local parser = xml.Xml(myxml)
        if not parser:is_valid() then
            logger.log("XRNI: Cannot parse xml file " .. parser:error() .. "...")
            return 0
        end

        local content = parser:convertToTable()

        if not xrni.table_contains(content, "RenoiseInstrument") then
            logger.log("XRNI: Cannot parse xml file, no <RenoiseInstrument> entry...")
            return 0
        end

        local instrument = content["RenoiseInstrument"]
        if not xrni.table_contains(instrument, "Samples") then
            logger.log("XRNI: Cannot parse xml file, no <Samples> entry...")
            return 0
        end

        inst_name = instrument["Name"]

        local samplesTable = instrument["Samples"]
        local samples = {}

        for i = 1, #samplesTable do
            local sample = samplesTable[i]

            samples[i] = {
                Name = sample["Name"] or ("Sample " .. tostring(i)),
                Volume = xrni.transform_volume(sample["Volume"]), -- 1
                Panning = xrni.transform_pan(sample["Panning"]), -- 0.5
                Transpose = xrni.transform_transpose(sample["Transpose"]), -- 0
                Finetune = xrni.transform_fine_tune(sample["Finetune"]), -- 0
                BeatSyncIsActive = xrni.transform_beat_sync_active(sample["BeatSyncIsActive"]), -- false
                BeatSyncLines = xrni.transform_beat_sync_lines(sample["BeatSyncLines"]), -- 16
                BaseNote = xrni.transform_note(sample["BaseNote"]), -- 48
                LoopMode = xrni.transform_loop_mode(sample["LoopMode"]), -- 0
                LoopStart = xrni.transform_loop_start(sample["LoopStart"]), -- 0
                LoopEnd = xrni.transform_loop_end(sample["LoopEnd"]), -- 0
                --LoopRelease
                --InterpolationMode
                --AutoSeek
                --IsAlias
                --SingleSliceTriggerEnabled
                --AutoFade
                --NewNoteAction
                NoteStart = 0,
                NoteEnd = 119,
                VelocityStart = 0,
                VelocityEnd = 127,
            }
        end

        if not xrni.table_contains(instrument, "SampleSplitMap") then
            logger.log("XRNI: Cannot parse xml file, no <SampleSplitMap> entry...")
            return 0
        end

        local splitMapTable = instrument["SampleSplitMap"]

        if not xrni.table_contains(splitMapTable, "NoteOnMappings") then
            logger.log("XRNI: Cannot parse xml file, no <NoteOnMappings> entry...")
            return 0
        end

        local noteOnMappingsTable = splitMapTable["NoteOnMappings"]

        for i = 1, #noteOnMappingsTable do
            local noteOnMapping = noteOnMappingsTable[i]
            local sampleIndex = tonumber(noteOnMapping["SampleIndex"]) + 1

            local sample = samples[sampleIndex]
            if sample ~= nil then
                sample["BaseNote"] = xrni.transform_note(noteOnMapping["BaseNote"])
                sample["NoteStart"] = xrni.transform_note(noteOnMapping["NoteStart"])
                sample["NoteEnd"] = xrni.transform_note(noteOnMapping["NoteEnd"])
                sample["VelocityStart"] = xrni.transform_velocity(noteOnMapping["VelocityStart"])
                sample["VelocityEnd"] = xrni.transform_velocity(noteOnMapping["VelocityEnd"])
                sample["UseEnvelopes"] = xrni.transform_boolean(noteOnMapping["UseEnvelopes"])
                sample["MapVelocityToVolume"] = xrni.transform_boolean(noteOnMapping["MapVelocityToVolume"])
            end
        end

        -- Error check
        if #samples == 0 then
            logger.log("XRNI: No samples found...")
            return 0
        end

        -- Load samples
        local prg = engine:get_program(program_index)
        prg:clear()
        prg:set_name(inst_name)

        for i = 1, #samples do
            local smpl = samples[i]

            local z = prg:add_zone()
            z:set_name(engine:extract_name(smpl.Name))

            engine:load_sample_buffer(z, zfile:unpack_file(sample_files[i]), true)

            z.loop_start = smpl.LoopStart
            z.loop_end = smpl.LoopEnd
            z.loop_mode = smpl.LoopMode
            z.num_cues = 0 -- force reset cues

            z.lo_input_range.midi_key = smpl.NoteStart
            z.lo_input_range.midi_vel = smpl.VelocityStart
            z.hi_input_range.midi_key = smpl.NoteEnd
            z.hi_input_range.midi_vel = smpl.VelocityEnd

            z.midi_root_key = smpl.BaseNote
            z.midi_coarse_tune = smpl.Transpose
            --z.midi_fine_tune = smpl.Finetune
            z.mp_gain = smpl.Volume
            z.mp_pan = smpl.Panning
        end

        logger.log("XRNI: Finished importing 1 program/s...")
        return 1
    end

    logger.log("XRNI: Cannot parse Instrument.xml...")
    return 0
end

--==============================================================================
-- Register Format with the Engine !

engine:register_format({
    wildcard="*.xrni",
    import=xrni_import_instrument,
    version=5,
})
